import z from 'zod';
import { makeLimitOffsetParser, queryLimitParser, queryOffsetParser } from '../common/parsers/queryParams';
import { ReadonlyNonEmptyArray } from 'fp-ts/ReadonlyNonEmptyArray';
import {
  SUBSCRIPTION_CHANNEL_API_LITERAL_TYPE,
  SUBSCRIPTION_ID_PARAM_NAME,
  SUBSCRIPTION_ORIGIN_TYPES,
  SUBSCRIPTION_STATUSES,
  SUBSCRIPTION_USER_API_LITERAL_TYPE,
} from './constants';
import { ApiSubscription } from './types';
import { cast } from '../utils/types';
import { ListResponseTrait } from '../response/types';
import { stripeSubscriptionIdParser } from '../stripe/parsers';
import { currencyAmountParser, currencyParser } from '../currencies/parsers';
import { UserId } from '../user/types';
import { ChannelId } from '../channel/types';

export const SEPARATOR = '::'; // just coincides with UUID_SEPARATOR

// "api src" - looks just like a normal graph src but is not graph src; abstractions are just similar
// (in truth, its because the older api already works like that but we'd like to both move from the graph terminology and keep compatibility)
const makeApiIdParser = <
  L extends string,
  IdMap extends {
    [key in L]: Id;
  },
  Id extends string
>(
  literals: ReadonlyNonEmptyArray<L>
) => {
  const makeType = (literal: L) =>
    z.object({
      type: z.literal(literal),
      id: z.string().transform(cast<IdMap[L]>()),
    });
  // to procure type
  const first = makeType(literals[0]);
  return z.preprocess(
    (uuid) => {
      // no throws in preprocess https://github.com/colinhacks/zod/issues/1460
      if (typeof uuid !== 'string') {
        console.error(`Expected string, got ${typeof uuid}: ${uuid}`);
        return uuid;
      }
      const [type, id] = uuid.split(SEPARATOR);
      return { type, id: id as IdMap[L] };
    },
    // trick zod https://github.com/colinhacks/zod/issues/1442
    z.discriminatedUnion('type', [first, ...(literals.slice(1).map(makeType) as [typeof first, ...typeof first[]])])
  );
};
const srcLiterals = [SUBSCRIPTION_USER_API_LITERAL_TYPE] as const;
export const subscriptionApiSrcCompositeParser = makeApiIdParser<
  typeof srcLiterals[number],
  {
    user: UserId;
  },
  UserId
>(srcLiterals);
const dstLiterals = [SUBSCRIPTION_CHANNEL_API_LITERAL_TYPE] as const;
export const subscriptionApiDstCompositeParser = makeApiIdParser<
  typeof dstLiterals[number],
  {
    channel: ChannelId;
  },
  ChannelId
>(dstLiterals);
export const subscriptionApiSrcDstCompositeParserOptional = z.object({
  src: subscriptionApiSrcCompositeParser.optional(),
  dst: subscriptionApiDstCompositeParser.optional(),
});
export const subscriptionApiSrcDstCompositeParser = z.object({
  src: subscriptionApiSrcCompositeParser,
  dst: subscriptionApiDstCompositeParser,
});

export const subscriptionListQueryParser = makeLimitOffsetParser(queryLimitParser, queryOffsetParser).merge(
  subscriptionApiSrcDstCompositeParserOptional
);

export const subscriptionListResponseParser = z.any(/*stub*/).transform(cast<ListResponseTrait<ApiSubscription>>());
export const subscriptionGetOneParamsParser = z.object({
  [SUBSCRIPTION_ID_PARAM_NAME]: stripeSubscriptionIdParser,
});
export const subscriptionRecurringParser = z.boolean();
export const subscriptionAmountParser = currencyAmountParser;
export const subscriptionUnitParser = currencyParser;
export const subscriptionApiAmountParser = z.object({
  amt: subscriptionAmountParser,
});
export const subscriptionAmountUnitParser = z
  .object({
    unit: subscriptionUnitParser,
  })
  .merge(subscriptionApiAmountParser);
export const subscriptionStatusParser = z.enum(SUBSCRIPTION_STATUSES);
export const subscriptionApiRecurringParser = z.object({
  recurring: subscriptionRecurringParser,
});
export const subscriptionRecurringAndStatusParser = z
  .object({
    status: subscriptionStatusParser,
  })
  .merge(subscriptionApiRecurringParser);
export const subscriptionOriginTypeParser = z.enum(SUBSCRIPTION_ORIGIN_TYPES);
// "api" subscription shape; not for use to parse real stripe subscriptions
export const subscriptionCreateInputParser = z
  .object({
    // // TODO camelcase; used?
    // origin_type: subscriptionOriginTypeParser,
    // // TODO camelcase; used?
    // origin_id: subscriptionOriginIdParser,
  })
  .merge(subscriptionApiSrcDstCompositeParser)
  .merge(subscriptionAmountUnitParser)
  .merge(subscriptionApiRecurringParser);
export const subscriptionCreateResponseParser = z.any(/*stub*/).transform(cast<ApiSubscription>());

export const subscriptionUpdateInputParser = subscriptionApiAmountParser.partial();
export const subscriptionUpdateParamsParser = subscriptionGetOneParamsParser;
export const subscriptionUpdateResponseParser = subscriptionCreateResponseParser;
export const subscriptionDeleteParamsParser = subscriptionGetOneParamsParser;
