import z from 'zod';
import {
  SUBSCRIPTION_CHANNEL_API_LITERAL_TYPE,
  SUBSCRIPTION_ORIGIN_TYPES,
  SUBSCRIPTION_PROVIDERS,
  SUBSCRIPTION_STATUSES,
  SUBSCRIPTION_USER_API_LITERAL_TYPE,
} from './constants';
import { StripeSubscriptionId } from '../stripe/types';
import { Subset } from '../utils/types';
import { Currency } from '../currencies/types';
import {
  SEPARATOR,
  subscriptionApiDstCompositeParser,
  subscriptionApiSrcCompositeParser,
  subscriptionCreateInputParser,
  subscriptionCreateResponseParser,
  subscriptionDeleteParamsParser,
  subscriptionListQueryParser,
  subscriptionListResponseParser,
  subscriptionUpdateInputParser,
  subscriptionUpdateParamsParser,
  subscriptionUpdateResponseParser,
} from './parsers';
import { UserId } from '../user/types';
import { ChannelIdOfType, ChannelType } from '../channel/types';

export type SubscriptionId = StripeSubscriptionId;
export type SubscriptionProvider = typeof SUBSCRIPTION_PROVIDERS[number];
export type SubscriptionStatus = typeof SUBSCRIPTION_STATUSES[number];
export type SubscriptionOriginType = typeof SUBSCRIPTION_ORIGIN_TYPES[number];
type SubscriptionOriginIdPerTypeMapping = {
  paypal: never; // PaypalSubscriptionId,
  stripe: StripeSubscriptionId;
};
// eslint-disable-next-line @typescript-eslint/no-unused-vars
type AssertSubscriptionOriginIdPerTypeMapping1 = Subset<
  keyof SubscriptionOriginIdPerTypeMapping,
  SubscriptionOriginType
>;
// eslint-disable-next-line @typescript-eslint/no-unused-vars
type AssertSubscriptionOriginIdPerTypeMapping2 = Subset<
  SubscriptionOriginType,
  keyof SubscriptionOriginIdPerTypeMapping
>;
export type SubscriptionOriginIdForType<T extends SubscriptionOriginType> = SubscriptionOriginIdPerTypeMapping[T];

// amt, status, unit, maybe client_secret, maybe origin_id
// backward compat with current frontend
export type ApiSubscription = {
  amt: number;
  unit: Currency;
  status: SubscriptionStatus;
  // a secret to follow thru with subscription payment process; user in Stripe; TODO move out of this type
  client_secret?: string;
  src: PassportIssueSrc;
  dst: PassportIssueDst;
};

export type ApiSubscriptionSrcComposite = z.infer<typeof subscriptionApiSrcCompositeParser>;
export type ApiSubscriptionDstComposite = z.infer<typeof subscriptionApiDstCompositeParser>;

export type SubscriptionListApiQuery = z.infer<typeof subscriptionListQueryParser>;
export type SubscriptionListApiResponse = z.infer<typeof subscriptionListResponseParser>; // questionable whether we want to encode channel type there; probably not
export type SubscriptionUpdateApiParams = z.infer<typeof subscriptionUpdateParamsParser>;
export type SubscriptionUpdateApiBody = z.infer<typeof subscriptionUpdateInputParser>;
export type SubscriptionUpdateApiResponse = z.infer<typeof subscriptionUpdateResponseParser>;
export type SubscriptionCreateApiBody = z.infer<typeof subscriptionCreateInputParser>;
export type SubscriptionCreateApiResponse = z.infer<typeof subscriptionCreateResponseParser>;
export type SubscriptionDeleteApiParams = z.infer<typeof subscriptionDeleteParamsParser>;

export type PassportIssueSrc = `${typeof SUBSCRIPTION_USER_API_LITERAL_TYPE}${typeof SEPARATOR}${UserId}`;
export type PassportIssueDst<T extends ChannelType = ChannelType> =
  `${typeof SUBSCRIPTION_CHANNEL_API_LITERAL_TYPE}${typeof SEPARATOR}${ChannelIdOfType<T>}`;

enum PassportIssueBrand {
  _ = '',
}

export type PassportIssueId = string & PassportIssueBrand;
export type WithPassportId = {
  id: PassportIssueId;
};
export type PassportCore = {
  src: PassportIssueSrc;
  dst: PassportIssueDst;
  issuedAt: number;
  expiresAt: number;
};
// this is inner subscription type; its meaning is different from payment provider's subscription entity
// 1) it is created at the point of payment provider's subscription is activated
// 2) it is deactivated at the point of payment provider's subscription is deactivated
// 3) src and dst are immutable
// 4) it is not deleted, but marked as inactive, for the reason of lookups from the payments, potentially; in that sense, it is a cache
// 5) its meaning is access control
// - with src, dst and issuedAt, we can historically lookup the access
export type Passport = PassportCore & WithPassportId;
